//! This contains the `mangle_headers` function and its helpers
//!
//! Mangling headers is necessary to normalize CSV column names before
//! deserialization can be done by Serde.
//! Field detection occurs in `detect_field`.
//!
//! As of lib 0.4.0, unrecognized headers are ignored and left untouched by mangling.
//! However, they are not passed through to any subsequent exports from plate-tool.

pub fn mangle_headers(data: &str) -> String {
    let (header, rows) = data.split_at(data.find('\n').unwrap());
    let fields = header.trim().split(',');
    let mut modified_headers: Vec<String> = Vec::new();
    for field in fields {
        if let Some(f) = detect_field(field) {
            modified_headers.push(f.to_string());
        } else {
            log::debug!("Unk field: {:?}", field)
        }
    }

    modified_headers.join(",") + "\n" + rows
}

fn detect_field(field: &str) -> Option<Field> {
    // NOTE: Don't return none! Consider refactor to -> Field later on.
    // Returning none means a field will be dropped, which requires the user to manually drop
    // columns from their picklist.
    match field.trim().to_lowercase() {
        x if x.contains("source") || x.contains("src") => match x {
            _ if x.contains("plate") || x.contains("barcode") => Some(Field::SourcePlate),
            _ if x.contains("well") => Some(Field::SourceWell),
            _ if x.contains("format") || x.contains("fmt") => Some(Field::SourceFormat),
            _ => Some(Field::Other(x)), // Retain unknown fields
        },
        x if x.contains("destination") || x.contains("dest") => match x {
            _ if x.contains("plate") || x.contains("barcode") => Some(Field::DestinationPlate),
            _ if x.contains("well") => Some(Field::DestinationWell),
            _ if x.contains("format") || x.contains("fmt") => Some(Field::DestinationFormat),
            _ => Some(Field::Other(x)), // Retain unknown fields
        },
        x if x.contains("volume") => Some(Field::Volume),
        x if x.contains("concentration") => Some(Field::Concentration),
        x => Some(Field::Other(x)), // Retain unknown fields
    }
}

enum Field {
    SourcePlate,
    DestinationPlate,
    SourceWell,
    DestinationWell,
    SourceFormat,
    DestinationFormat,
    Volume,
    Concentration,
    Other(String),
}

impl ToString for Field {
    fn to_string(&self) -> String {
        match self {
            Field::SourcePlate => "sourceplate".to_string(),
            Field::DestinationPlate => "destinationplate".to_string(),
            Field::SourceWell => "sourcewell".to_string(),
            Field::DestinationWell => "destinationwell".to_string(),
            Field::SourceFormat => "sourceformat".to_string(),
            Field::DestinationFormat => "destinationformat".to_string(),
            Field::Volume => "volume".to_string(),
            Field::Concentration => "concentration".to_string(),
            Field::Other(x) => x.to_string(),
        }
    }
}
